"""Base Scheduler class."""

from abc import ABC, abstractclassmethod

from blackcap.schemas.api.schedule.post import ScheduleCreate
from blackcap.schemas.schedule import Schedule


class BaseScheduler(ABC):
    """Base Scheduler class."""

    CONFIG_KEY = "SCHEDULER"
    CONFIG_KEY_DEF_VAL = "RANDOM"

    # Change this value in custom auther implementations.
    CONFIG_KEY_VAL = "RANDOM"

    @abstractclassmethod
    def schedule(
        self: "BaseScheduler", schedule_create: ScheduleCreate
    ) -> Schedule:  # noqa: E501
        """Create schedule from schedule request.

        Args:
            schedule_create (ScheduleCreate): Schedule create request

        Returns:
            Schedule: Instance of Schedule
        """
        pass



"""User BLoCs."""

from typing import List

from blackcap.configs import config_registry
from blackcap.db import DBSession
from blackcap.auther import auther_registry
from blackcap.models.protagonist import ProtagonistDB
from blackcap.schemas.api.user.delete import UserDelete
from blackcap.schemas.api.user.get import UserGetQueryParams, UserQueryType
from blackcap.schemas.api.user.post import UserCreate
from blackcap.schemas.api.user.put import UserUpdate
from blackcap.schemas.user import User

from logzero import logger

from sqlalchemy import select


config = config_registry.get_config()
auther = auther_registry.get_auther(config.AUTHER)


def create_user(user_create_list: List[UserCreate]) -> List[User]:
    """Create user and save to DB from UserCreate request.

    Args:
        user_create_list (List[UserCreate]): List of UserCreate request

    Raises:
        Exception: error

    Returns:
        List[User]: Created user
    """
    try:
        user_list = auther.register_user(user_create_list)
        return user_list
    except Exception as e:
        logger.error(f"Unable to register users due to {e}")
        raise e


def get_users(query_params: UserGetQueryParams) -> List[User]:
    """Query DB for users.

    Args:
        query_params (UserGetQueryParams): User Query params

    Raises:
        Exception: error

    Returns:
        List[User]: List of Users
    """
    user_list: List[ProtagonistDB] = []

    if query_params.query_type == UserQueryType.GET_ALL_USERS:
        stmt = select(ProtagonistDB)
    if query_params.query_type == UserQueryType.GET_USERS_BY_ID:
        stmt = select(ProtagonistDB).where(
            ProtagonistDB.id == query_params.user_id
        )  # noqa: E501
    if query_params.query_type == UserQueryType.GET_USERS_BY_EMAIL:
        stmt = select(ProtagonistDB).where(
            ProtagonistDB.email == query_params.email
        )  # noqa: E501
    if query_params.query_type == UserQueryType.GET_USERS_BY_ORGANISATION:
        stmt = select(ProtagonistDB).where(
            ProtagonistDB.organisation == query_params.organisation
        )

    with DBSession() as session:
        try:
            user_list: List[ProtagonistDB] = (
                session.execute(stmt).scalars().all()
            )  # noqa: E501
            user_list = [
                User(user_id=obj.id, **obj.to_dict()) for obj in user_list
            ]  # noqa: E501
            return user_list
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to fetch users due to {e}")
            raise e


def update_user(user_update: UserUpdate) -> User:
    """Update user in the DB from UserUpdate request.

    Args:
        user_update (UserUpdate): User update request

    Returns:
        User: Instance of updated user
    """
    stmt = select(ProtagonistDB).where(ProtagonistDB.id == user_update.user_id)
    with DBSession() as session:
        try:
            user_list: List[ProtagonistDB] = (
                session.execute(stmt).scalars().all()
            )  # noqa: E501
            if not user_list:
                # TODO: raise not found error
                pass
            if len(user_list) == 1:
                user_update_dict = user_update.dict()
                user_update_dict.pop("user_id")
                updated_user = user_list[0].update(
                    session, **user_update.dict()
                )  # noqa: E501
                return User(user_id=updated_user.id, **updated_user.to_dict())
        except Exception as e:
            session.rollback()
            logger.error(
                f"Unable to update user: {user_update.dict()} due to {e}"
            )  # noqa: E501
            # TODO: Raise error


def delete_user(user_delete: UserDelete) -> User:
    """Delete user in the DB from UserDelete request.

    Args:
        user_delete (UserDelete): User delete request

    Raises:
        Exception: error

    Returns:
        User: Instance of deleted user
    """
    stmt = select(ProtagonistDB).where(ProtagonistDB.id == user_delete.user_id)
    with DBSession() as session:
        try:
            user_list: List[ProtagonistDB] = (
                session.execute(stmt).scalars().all()
            )  # noqa: E501
            if not user_list:
                # TODO: raise not found error
                pass
            if len(user_list) == 1:
                deleted_user = user_list[0].delete(session)
                return User(user_id=deleted_user.id, **deleted_user.to_dict())
        except Exception as e:
            session.rollback()
            logger.error(
                f"Unable to delete user: {user_delete.dict()} due to {e}"
            )  # noqa: E501
            raise e



"""Blackcap common schemas."""

from typing import Any, List

from pydantic import BaseModel


class ResponseSchema(BaseModel):
    """API response schema."""

    msg: str
    items: List[Any] = []
    errors: List[Any] = []



"""GCP implementation of messenger."""

import json
from typing import Callable, Dict, Optional

from blackcap.db import DBSession
from blackcap.configs.base import BaseConfig
from blackcap.messenger.base import BaseMessenger
from blackcap.models.schedule import ScheduleDB
from blackcap.schemas.message import Message, MessageType
from blackcap.schemas.schedule import Schedule
from blackcap.utils.json_encoders import UUIDEncoder

from google.auth import jwt
from google.cloud import pubsub_v1
from google.cloud.pubsub_v1.subscriber.message import Message as GCPMessage

from logzero import logger

from sqlalchemy import select


class GCPMessenger(BaseMessenger):
    """GCP(Pub/Sub) implementation of Messenger."""

    CONFIG_KEY_VAL = "GCP"

    def __init__(self: "GCPMessenger", config: BaseConfig) -> None:
        """Initialize Messenger with app config.

        Args:
            config (BaseConfig): Config to initialize messenger
        """
        self.pub_audience = (
            "https://pubsub.googleapis.com/google.pubsub.v1.Publisher"  # noqa: E501
        )
        self.sub_audience = (
            "https://pubsub.googleapis.com/google.pubsub.v1.Subscriber"  # noqa: E501
        )
        self.service_account_info = json.load(
            open(config.GOOGLE_APPLICATION_CREDENTIALS)
        )  # noqa: E501
        self.pub_credentials = jwt.Credentials.from_service_account_info(
            self.service_account_info, audience=self.pub_audience
        )

        self.sub_credentials = jwt.Credentials.from_service_account_info(
            self.service_account_info, audience=self.sub_audience
        )

        self.project_id = config.GCP_PROJECT_ID

    @property
    def publisher(self) -> pubsub_v1.PublisherClient:
        """Publisher Client."""
        return pubsub_v1.PublisherClient(credentials=self.pub_credentials)

    @property
    def subscriber(self) -> pubsub_v1.SubscriberClient:
        """Subscriber Client."""
        return pubsub_v1.SubscriberClient(credentials=self.sub_credentials)

    def publish(self: "GCPMessenger", msg: Dict, topic_id: str) -> str:
        """Publish msg on the GCP Pub/Sub queue.

        Args:
            msg (Dict): Messsag to publish
            topic_id (str): Id of the topic

        Returns:
            str: Id of the published msg
        """
        topic_path = self.publisher.topic_path(self.project_id, topic_id)

        # Msg must be a bytestring

        msg = json.dumps(msg, cls=UUIDEncoder).encode("utf-8")

        # Wait for the returned future
        future = self.publisher.publish(topic_path, msg)
        return future.result()

    def subscribe(
        self: "GCPMessenger",
        callback: Callable,
        sub_id: str,
        timeout: Optional[float] = None,  # noqa: E501
    ) -> None:
        """Subscribe to a topic.

        Args:
            callback (Callable): Callback to invoke when a msg is received
            sub_id (str): Id of the topic.
            timeout (Union[float, None]): Time to wait for msgs. Defaults to None. # noqa: E501
        """
        subscription_path = self.subscriber.subscription_path(
            self.project_id, sub_id
        )  # noqa: E501

        streaming_pull_future = self.subscriber.subscribe(
            subscription_path,
            callback=callback,
            await_callbacks_on_shutdown=True,  # noqa: E501
        )

        with self.subscriber:
            try:
                streaming_pull_future.result(timeout=timeout)
            except TimeoutError as e:
                logger.error(
                    f"GCPMessenger timeout error while pulling messages. Error: {e}"  # noqa: E501
                )  # noqa: E501
                streaming_pull_future.cancel()

    def process_schedule_msg(self: "GCPMessenger", msg: GCPMessage) -> None:
        """Process Schedule Pub/Sub msgs.

        Args:
            msg (GCPMessage): Pub/Sub Msg
        """
        parsed_msg = Message.parse_raw(msg.data)
        if (
            parsed_msg.msg_type
            == MessageType.TO_CONDUCTOR_JOB_STATUS_UPDATE.value  # noqa: E501
        ):
            logger.info(f"Recieved msg: {parsed_msg.dict()}")
            schedule = Schedule(**parsed_msg.data)
            # Check if job already exists
            stmt = select(ScheduleDB).where(
                ScheduleDB.id == schedule.schedule_id
            )  # noqa: E501
            with DBSession() as session:
                fetched_schedules = session.execute(stmt).scalars().all()
                if len(fetched_schedules) == 0:
                    logger.error(
                        f"""
                    Unable to find job from schedule!!!
                    Schedule ID: {schedule.id},
                    Job ID: {schedule.job_id}
                    """
                    )
                else:
                    # Update schedule status
                    fetched_schedules[0].update(
                        session, status=schedule.status
                    )  # noqa: E501
            msg.ack()

    def echo_msg(self: "GCPMessenger", msg: Message) -> None:
        """Echo msgs to stdout.

        Args:
            msg (Message): Message to echo
        """
        print(msg)
        msg.ack()



"""Blackcap Schedule DELETE route schemas."""

from pydantic import BaseModel
from pydantic.types import UUID4


class ScheduleDelete(BaseModel):
    """Schema to parse delete schedule requests."""

    schedule_id: UUID4



"""Random Scheduler implementation of Scheduler."""


from typing import List

from blackcap.observer.base import BaseObserver
from blackcap.schemas.metrics import Metrics


class ElasticObserver(BaseObserver):
    """Elastic observer to fetch metrics from Elastic stack."""

    CONFIG_KEY_VAL = "ELASTIC"

    def get_metrics(self: "BaseObserver", range: str) -> List[Metrics]:
        """Get cluster metrics.

        Args:
            range (str): time range

        Returns:
            List[Metrics]: List of metrics from each cluster # noqa: DAR202
        """
        return []



"""Observer to fetch metrcis from a central monitoring system."""

from blackcap.observer.elastic_observer import ElasticObserver
from blackcap.observer.registry import ObserverRegistry


observer_registry = ObserverRegistry()
observer_registry.add_observer(ElasticObserver())



"""Blackcap workers."""

from blackcap.configs import config_registry

from celery import Celery

celery_app = Celery(
    __name__,
    broker=config_registry.get_config().CELERY_BROKER_URL,
    backend=config_registry.get_config().CELERY_RESULT_BACKEND,
    include=["blackcap.tasks"],
)

celery_app.conf.task_routes = {"blackcap.tasks.*": {"queue": "blackcap"}}



"""Server factory."""

from typing import Callable

from blackcap import celery_app
from blackcap.configs.base import BaseConfig
from blackcap.routes.job import job_bp
from blackcap.routes.schedule import schedule_bp
from blackcap.routes.status import status_bp
from blackcap.routes.user import user_bp
from blackcap.tasks import init_celery

from flask import Flask

from logzero import logger


def register_extensions(app: Flask) -> None:
    """Register extension.

    Args:
        app (Flask): Flask app
    """
    init_celery(app, celery_app)
    logger.info("Registered extensions")


def register_blueprints(app: Flask) -> None:
    """Register blueprints.

    Args:
        app (Flask): Flask app
    """
    app.register_blueprint(status_bp)
    app.register_blueprint(job_bp)
    app.register_blueprint(user_bp)
    app.register_blueprint(schedule_bp)
    logger.info("Registered blueprints")


def create_app(
    config: BaseConfig,
    register_ext: Callable[[Flask], None] = register_extensions,
    register_bp: Callable[[Flask], None] = register_blueprints,
) -> Flask:
    """Create a Flask App.

    Args:
        config (BaseConfig): App config
        register_ext (Callable[[Flask], None]): Function to register extensions
        register_bp (Callable[[Flask], None]): Function to register Blueprints

    Returns:
        Flask: Flask app
    """
    logger.info("Creating a flask app...")
    app = Flask(__name__)
    logger.info(f"Flask app created: {app.import_name}")

    # Load config
    app.config.from_object(config)
    logger.info(f"Config loaded: {type(config).__name__}")

    # Register extension
    register_ext(app)

    # Register blueprints
    register_bp(app)

    return app



"""Cluster DBModel."""


from blackcap.models.meta.mixins import (
    DBModel,
    SurrogatePKUUID,
    TimestampMixin,
)  # noqa: E501


from sqlalchemy import Column, String


class ClusterDB(DBModel, TimestampMixin, SurrogatePKUUID):
    """Cluster table."""

    __tablename__ = "cluster"
    name = Column(String(), nullable=False)
    cluster_type = Column(String(), nullable=False)
    # TODO: this should be an enum with the possibles status valiue (online, offline, error... whatever)  # noqa: E501
    status = Column(String(), nullable=False)
    cluster_caps = Column(String(), nullable=True)
    messenger = Column(String(), nullable=False, default="GCP")
    messenger_queue = Column(String(), nullable=False)



"""Job BLoCs."""

from typing import List

from blackcap.db import DBSession
from blackcap.models.job import JobDB
from blackcap.schemas.api.job.delete import JobDelete
from blackcap.schemas.api.job.get import JobGetQueryParams, JobQueryType
from blackcap.schemas.api.job.post import JobCreate
from blackcap.schemas.api.job.put import JobUpdate
from blackcap.schemas.job import Job

from logzero import logger

from sqlalchemy import select


def create_job(job_create_list: List[JobCreate]) -> List[Job]:
    """Create jobs in the DB from JobCreate request.

    Args:
        job_create_list (List[JobCreate]): JobCreate request

    Raises:
        Exception: error

    Returns:
        List[Job]: List of created clusters
    """
    with DBSession() as session:
        try:
            job_db_create_list: List[JobDB] = [
                JobDB(
                    protagonist_id=job_create.user.user_id,
                    **job_create.job.dict(exclude={"job_id"}),  # noqa: E501
                )
                for job_create in job_create_list
            ]
            JobDB.bulk_create(job_db_create_list, session)
            return [
                Job(job_id=obj.id, **obj.to_dict())
                for obj in job_db_create_list  # noqa: E501
            ]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to create jobs: {e}")
            raise e


def get_jobs(query_params: JobGetQueryParams) -> List[Job]:
    """Query DB for jobs.

    Args:
        query_params (JobGetQueryParams): Query params from request

    Raises:
        Exception: error

    Returns:
        List[Job]: List of jobs returned from DB
    """
    job_list: List[JobDB] = []

    stmt = ""

    if query_params.query_type == JobQueryType.GET_ALL_JOBS:
        stmt = select(JobDB)
    if query_params.query_type == JobQueryType.GET_JOBS_BY_ID:
        stmt = select(JobDB).where(JobDB.id == query_params.job_id)
    if query_params.query_type == JobQueryType.GET_JOBS_BY_CLUSTER_ID:
        stmt = select(JobDB).where(JobDB.id == query_params.cluster_id)
    if query_params.query_type == JobQueryType.GET_JOBS_BY_PROTAGONIST_ID:
        stmt = select(JobDB).where(JobDB.id == query_params.protagonist_id)
    if query_params.query_type == JobQueryType.GET_JOBS_BY_STATUS:
        stmt = select(JobDB).where(JobDB.id == query_params.job_status)

    with DBSession() as session:
        try:
            job_list: List[JobDB] = session.execute(stmt).scalars().all()  # noqa: E501
            return [Job(job_id=obj.id, **obj.to_dict()) for obj in job_list]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to fetch jobs due to {e}")
            raise e

    return job_list


def update_job(job_update: JobUpdate) -> Job:
    """Update job in the DB from JobUpdate request.

    Args:
        job_update (JobUpdate): JobUpdate request

    Raises:
        Exception: error

    Returns:
        Job: Instance of Updated Job
    """
    stmt = select(JobDB).where(JobDB.id == job_update.job_id)
    with DBSession() as session:
        try:
            job_list: List[JobDB] = session.execute(stmt).scalars().all()  # noqa: E501
            if len(job_list) == 1:
                job_update_dict = job_update.dict()
                job_update_dict.pop("job_id")
                updated_job = job_list[0].update(
                    session, **job_update.dict()
                )  # noqa: E501
                return Job(job_id=updated_job.id, **updated_job.to_dict())
            if len(job_list) == 0:
                # TODO: Raise not found
                pass
        except Exception as e:
            session.rollback()
            logger.error(
                f"Unable to update job: {job_update.dict()} due to {e}"
            )  # noqa: E501
            raise e


def delete_job(job_delete: JobDelete) -> Job:
    """Delete job in the DB from JobDelete request.

    Args:
        job_delete (JobDelete): JobDelete request

    Raises:
        Exception: error

    Returns:
        Job: Instance of Deleted Job
    """
    stmt = select(JobDB).where(JobDB.id == job_delete.job_id)
    with DBSession() as session:
        try:
            job_list: List[JobDB] = session.execute(stmt).scalars().all()  # noqa: E501
            if len(job_list) == 1:
                deleted_job = job_list[0].delete(session)
                return Job(job_id=deleted_job.id, **deleted_job.to_dict())
            if len(job_list) == 0:
                # TODO: Raise not found
                pass
        except Exception as e:
            session.rollback()
            logger.error(
                f"Unable to delete job: {job_delete.dict()} due to {e}"
            )  # noqa: E501
            raise e



"""Blackcap User GET route schemas."""

from enum import Enum, unique
from typing import List, Optional

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.user import User

from pydantic import BaseModel


@unique
class UserQueryType(Enum):
    """User Query type enum."""

    GET_ALL_USERS = "get_all_users"
    GET_USERS_BY_ID = "get_user_by_id"
    GET_USERS_BY_EMAIL = "get_users_by_email"
    GET_USERS_BY_ORGANISATION = "get_users_by_organisation"


class UserGetQueryParams(BaseModel):
    """User GET request query params schema."""

    query_type: UserQueryType
    user_id: Optional[str]
    email: Optional[str]
    organisation: Optional[str]


class UserGetResponse(ResponseSchema):
    """User GET response schema."""

    items: List[User] = []



"""Blackcap configs."""
from enum import Enum, unique
from functools import lru_cache

from blackcap.configs.default import DefaultConfig
from blackcap.configs.registry import ConfigRegistry

config_registry = ConfigRegistry()
config_registry.add_config(DefaultConfig())



"""Blackcap Auth POST route schemas."""

from typing import List

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.user import User

from pydantic import BaseModel


class AuthUserCreds(BaseModel):
    """Schema to parse User credentials."""

    email: str
    password: str


class AuthPOSTResponse(ResponseSchema):
    """User POST response schema."""

    items: List[User] = []



"""Blackcap."""

from importlib.metadata import PackageNotFoundError, version  # type: ignore

from blackcap import auther  # noqa: F401
from blackcap import blocs  # noqa: F401
from blackcap import cli  # noqa: F401
from blackcap import configs  # noqa: F401
from blackcap import messenger  # noqa: F401
from blackcap import models  # noqa: F401
from blackcap import observer  # noqa: F401
from blackcap import routes  # noqa: F401
from blackcap import scheduler  # noqa: F401
from blackcap import schemas  # noqa: F401

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



"""Blackcap Cluster DELETE route schemas."""

from pydantic import BaseModel
from pydantic.types import UUID4


class ClusterDelete(BaseModel):
    """Schema to parse delete cluster requests."""

    cluster_id: UUID4



"""Cookie Auther implementation of Auther."""

from datetime import datetime, timedelta
from typing import List, Optional, Tuple

from bcrypt import checkpw, gensalt, hashpw

from blackcap.db import DBSession
from blackcap.auther.base import BaseAuther
from blackcap.configs import config_registry
from blackcap.models.protagonist import ProtagonistDB
from blackcap.schemas.api.auth.post import AuthUserCreds
from blackcap.schemas.api.user.post import UserCreate
from blackcap.schemas.user import User

from flask.wrappers import Request

import jwt

from logzero import logger

from sqlalchemy import select

config = config_registry.get_config()


class CookieAuther(BaseAuther):
    """Cookie Auther."""

    CONFIG_KEY_VAL = "COOKIE"

    def register_user(
        self: "BaseAuther", user_create_list: List[UserCreate]
    ) -> List[User]:
        """Register user.

        Args:
            user_create_list (List[UserCreate]): List of users to register

        Raises:
            Exception: error

        Returns:
            List(User): List of registered users
        """
        with DBSession() as session:
            try:
                protagonist_objs = [
                    ProtagonistDB(
                        password=hashpw(
                            user_create.password.encode(), gensalt()
                        ),  # noqa:E501
                        **user_create.user.dict(exclude={"user_id"}),
                    )
                    for user_create in user_create_list
                ]
                ProtagonistDB.bulk_create(protagonist_objs, session)
                return [
                    User(user_id=obj.id, **obj.to_dict())
                    for obj in protagonist_objs  # noqa: E501
                ]
            except Exception as e:
                session.rollback()
                logger.error(f"Unable to register users due to {e}")
                raise e

    def login_user(
        self: "BaseAuther", user_creds: AuthUserCreds
    ) -> Optional[Tuple[User, str]]:
        """Login user.

        Args:
            user_creds (AuthUserCreds): user creds

        Raises:
            Exception: error

        Returns:
            Optional[Tuple[User, str]]: Tuple containing user and cookie or None  # noqa: E501
        """
        with DBSession() as session:
            try:
                # Find user
                stmt = select(ProtagonistDB).where(
                    ProtagonistDB.email == user_creds.email
                )
                user_list: List[ProtagonistDB] = (
                    session.execute(stmt).scalars().all()
                )  # noqa: E501
                # return early if user not found
                if len(user_list) == 0:
                    return None

                # Check password hash
                if checkpw(
                    user_creds.password.encode(), user_list[0].password
                ):  # noqa: E501
                    # Create cookie and return user and cookie
                    user_cookie = jwt.encode(
                        {
                            "exp": datetime.utcnow()
                            + timedelta(
                                minutes=config.ACCESS_TOKEN_EXPIRE_MINUTES  # noqa: E501
                            ),
                            "sub": user_list[0].email,
                            "email": user_list[0].email,
                            "name": user_list[0].name,
                            "organisation": user_list[0].organisation,
                        },
                        config.SECRET_KEY,
                    )
                    return (
                        User(
                            user_id=user_list[0].id, **user_list[0].to_dict()
                        ),  # noqa: E501
                        user_cookie,
                    )
                else:
                    return None
            except Exception as e:
                session.rollback()
                logger.error(f"Unable to login user due to: {e}")
                raise e

    def logout_user(self: "BaseAuther") -> None:
        """Logout user."""
        pass

    def extract_user_from_token(
        self: "BaseAuther", token: str
    ) -> Optional[User]:  # noqa: E501
        """Extract user from token.

        Args:
            token (str): user access token

        Raises:
            Exception: error

        Returns:
            Optional[User]: Instance of User or None
        """
        with DBSession() as session:
            user_cookie_encoded = token
            # return early if cookie not found in request
            if user_cookie_encoded is None:
                return None
            # try to decode the cookie
            try:
                user_cookie_decoded = jwt.decode(
                    user_cookie_encoded,
                    config.SECRET_KEY,
                    algorithms=["HS256"],  # noqa: E501
                )
            except jwt.DecodeError as e:
                # return none if failed to decode cookie
                logger.error(f"Unable to decode cookie due to: {e}")
                return None
            except Exception as e:
                logger.error(f"Unable to decode cookie due to: {e}")
                raise e

            # Fetch and return user if present or None
            stmt = select(ProtagonistDB).where(
                ProtagonistDB.email == user_cookie_decoded["email"]
            )
            user_list: List[ProtagonistDB] = (
                session.execute(stmt).scalars().all()
            )  # noqa: E501
            # return early if user not found
            if len(user_list) == 0:
                return None
            else:
                return User(user_id=user_list[0].id, **user_list[0].to_dict())

    def authorize_user(
        self: "BaseAuther", user: User, request: Request
    ) -> bool:  # noqa: E501
        """Authorize user actions on resources.

        Args:
            user (User): Instance of User
            request (Request): Instance of Flask Request

        Returns:
            bool: Authorization decision
        """
        return True



"""Protagonist DBModel."""


from blackcap.models.meta.mixins import (
    DBModel,
    SurrogatePKUUID,
    TimestampMixin,
)  # noqa: E501

from sqlalchemy import Boolean, Column, LargeBinary, String


class ProtagonistDB(DBModel, TimestampMixin, SurrogatePKUUID):
    """Protagonist table."""

    __tablename__ = "protagonist"
    name = Column(String(), nullable=False)
    organisation = Column(String(), nullable=True)
    email = Column(String(), unique=True, nullable=False)
    # Hash password before storing it here
    password = Column(LargeBinary(128), nullable=True)
    active = Column(Boolean(), default=False)
    roles = Column(String(), nullable=True, default="USER")



"""Schedule commands."""
# flake8: noqa: DAR101

from pathlib import Path

import click

from blackcap.blocs.job import create_job
from blackcap.schemas.api.job.post import JobCreate
from blackcap.schemas.job import Job

from logzero import logger


@click.command()
@click.option("--file", help="file path of job's JSON file", required=True)
def random(file: str) -> None:
    """Schedule a job."""
    logger.info("Trying to schedule the job...")
    try:
        job_file_path = Path(file)
        with open(job_file_path) as job_file:
            data = job_file.read()
            job = Job.parse_raw(data)
            job_create = JobCreate(job=job)
            created_job = create_job(job_create)
            logger.info(f"\nSuccess!\nJob: {created_job}")
    except Exception as e:
        logger.error(f"failed to schedule the job: {e}")
        raise e


@click.group()
def sched() -> None:
    """Schedule jobs."""
    pass


sched.add_command(random)



"""Helper ORM functions."""

from typing import Dict

from sqlalchemy import Column, ForeignKey


def reference_col(
    tablename: str,
    nullable: bool = False,
    pk_name: str = "id",
    foreign_key_kwargs: Dict = None,
    column_kwargs: Dict = None,
) -> Column:
    """Column that adds primary key foreign key reference.

    Args:
        tablename (str): name of the table.
        nullable (bool): is nullable?
        pk_name (str): name of the primary key field.
        foreign_key_kwargs (Dict): Dict of args for fk.
        column_kwargs (Dict): Dict of args for column.

    Returns:
        Return the Column.

    Examples: ::
        category_id = reference_col('category')
        category = relationship('Category', backref='categories')
    """
    foreign_key_kwargs = foreign_key_kwargs or {}
    column_kwargs = column_kwargs or {}

    return Column(
        ForeignKey(f"{tablename}.{pk_name}", **foreign_key_kwargs),
        nullable=nullable,
        **column_kwargs,
    )



"""Base messenger class."""

from abc import ABC, abstractclassmethod
from typing import Callable, Dict, Optional


class BaseMessenger(ABC):
    """Base messenger class."""

    CONFIG_KEY = "MESSENGER"
    CONFIG_KEY_DEF_VAL = "GCP"

    # Change this value in custom auther implementations.
    CONFIG_KEY_VAL = "GCP"

    @abstractclassmethod
    def publish(self: "BaseMessenger", msg: Dict, topic_id: str) -> str:  # noqa: E501
        """Publish a msg.

        Args:
            msg (Dict): Message to publish
            topic_id (str): Id of the topic
        """
        pass

    @abstractclassmethod
    def subscribe(
        self: "BaseMessenger",
        callback: Callable,
        sub_id: str,
        timeout: Optional[float] = None,
    ) -> None:  # noqa: E501
        """Subscribe to a topic.

        Args:
            callback (Callable): Callback to invoke when a msg is received
            sub_id (str): Id of the topic.
            timeout (Union[float, None]): Time to wait for msgs. Defaults to None. # noqa: E501
        """
        pass



"""subscribe commands."""
# flake8: noqa: DAR101

import click

from blackcap.configs import config_registry
from blackcap.messenger import messenger_registry

from logzero import logger

config = config_registry.get_config()
messenger = messenger_registry.get_messenger(config.MESSENGER)


@click.command()
@click.option("--sub_id", default=config.GCP_PUBSUB_SUB_ID, help="Subscription Id")
def echo(sub_id: str) -> None:
    """subscribe and echo msgs"""
    logger.info("Trying to subscribe and read messages...")
    try:
        messenger.subscribe(callback=messenger.echo_msg, sub_id=sub_id)
    except Exception as e:
        logger.error(f"failed to subscribe: {e}")


@click.command()
@click.option("--sub_id", default=config.GCP_PUBSUB_SUB_ID, help="Subscription Id")
def schedule(sub_id: str) -> None:
    """subscribe and save msgs to DB"""
    logger.info("Trying to subscribe and read messages...")
    try:
        messenger.subscribe(callback=messenger.process_schedule_msg, sub_id=sub_id)
    except Exception as e:
        logger.error(f"failed to subscribe: {e}")


@click.group()
def sub() -> None:
    """GCP pub/sub channel subscribe commands."""
    pass


sub.add_command(echo)
sub.add_command(schedule)



"""Blackcap User POST route schemas."""

from typing import List

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.user import User

from pydantic import BaseModel


class UserCreate(BaseModel):
    """Schema to parse create User requests."""

    user: User
    password: str


class UserPOSTRequest(BaseModel):
    """User POST request schema."""

    users: List[UserCreate]


class UserPOSTResponse(ResponseSchema):
    """User POST response schema."""

    items: List[User] = []



"""Blackcap Job POST route schemas."""

from typing import List, Optional

from blackcap.schemas.job import Job
from blackcap.schemas.user import User

from pydantic import BaseModel


class JobPOSTResponse(BaseModel):
    """Job POST response schema."""

    items: List[Job] = []


class JobCreate(BaseModel):
    """Schema to parse create job requests."""

    job: Job
    user: Optional[User]


class JobPOSTRequest(BaseModel):
    """Job POST request schema."""

    jobs: List[JobCreate]



"""Auther to handle authentication and authorizations."""

from blackcap.auther.cookie_auther import CookieAuther
from blackcap.auther.registry import AutherRegistry


auther_registry = AutherRegistry()
# Add auther implementations to the registry
auther_registry.add_auther(CookieAuther())



"""Blackcap user routes."""


from flask import Blueprint

user_bp = Blueprint("user", __name__, url_prefix="/v1/user")

from blackcap.routes.user.get import get_user  # noqa: F401, E402, I100
from blackcap.routes.user.post import post_user  # noqa: F401, E402, I100



"""Job DBModel."""


from blackcap.models.meta.mixins import (
    DBModel,
    SurrogatePKUUID,
    TimestampMixin,
)  # noqa: E501
from blackcap.models.meta.orm import reference_col

from sqlalchemy import Column, Integer, String

# from sqlalchemy.orm import relationship


class JobDB(DBModel, TimestampMixin, SurrogatePKUUID):
    """Job table."""

    __tablename__ = "job"
    serialize_rules = ("-protagonist.jobs",)
    name = Column(String(), nullable=False)
    script = Column(String(), nullable=False)

    cluster_caps_req = Column(String(), nullable=True)

    protagonist_id = reference_col("protagonist")
    # protagonist = relationship("ProtagonistDB", backref="jobs")

    # it's going ot be more performance to keep a counter
    # on this table than queries the schedules count each time
    # we can create a db trigger to keep this updated
    # TODO: add a signal to keep this value updated
    schedules_count = Column(Integer, nullable=False, default=0)



"""Blackcap create commands."""
# flake8: noqa

from pathlib import Path
import click

from blackcap.auther import auther_registry
from blackcap.blocs.cluster import create_cluster
from blackcap.blocs.job import create_job, get_jobs, JobGetQueryParams, JobQueryType
from blackcap.blocs.schedule import create_schedule
from blackcap.blocs.user import create_user
from blackcap.configs import config_registry
from blackcap.schemas.api.cluster.post import ClusterCreate
from blackcap.schemas.api.job.post import JobCreate
from blackcap.schemas.api.schedule.post import ScheduleCreate
from blackcap.schemas.api.user.post import UserCreate
from blackcap.schemas.cluster import Cluster
from blackcap.schemas.job import Job
from blackcap.schemas.schedule import Schedule
from blackcap.schemas.user import User

config = config_registry.get_config()


@click.command()
@click.option("--email", required=True, help="email of the user")
@click.option("--password", required=True, help="password of the user")
@click.option("--name", default="CLI User", help="name of the user")
@click.option("--org", default="EBI", help="organisation of the user")
def user(email, password, name, org) -> None:
    """Create a user."""
    user_create_request = UserCreate(
        user=User(name=name, email=email, organisation=org), password=password
    )
    created_user_list = create_user([user_create_request])
    click.secho(f"User successfully created!\n\n{created_user_list[0]}", fg="green")


@click.command()
@click.option("--name", required=True, help="name of the cluster")
@click.option("--cluster_type", required=True, help="type of cluster")
@click.option("--cluster_caps", default="CPU", help="caps of cluster")
@click.option("--messenger", default="GCP", help="messenger the cluster uses")
@click.option("--queue", default="test-topic", help="messenger topic the cluster uses")
def cluster(name, cluster_type, cluster_caps, messenger, queue) -> None:
    """Create cluster."""
    # fetch a user first
    user_access_token = config.USER_ACCESS_TOKEN
    auther = auther_registry.get_auther(config.AUTHER)
    user = auther.extract_user_from_token(user_access_token)

    cluster_create_request = ClusterCreate(
        cluster=Cluster(
            name=name,
            cluster_type=cluster_type,
            cluster_caps=cluster_caps,
            messenger=messenger,
            messenger_queue=queue,
        ),
        user=user,
    )

    created_cluster_list = create_cluster([cluster_create_request])
    click.secho(
        f"Cluster successfully created!\n\n{created_cluster_list[0]}", fg="green"
    )


@click.command()
@click.option("--name", required=True, help="name of the job")
@click.option("--script", required=True, help="path of the script")
def job(name, script) -> None:
    """Create cluster."""
    # fetch a user first
    user_access_token = config.USER_ACCESS_TOKEN
    auther = auther_registry.get_auther(config.AUTHER)
    user = auther.extract_user_from_token(user_access_token)
    script_path = Path(script)
    with open(script_path) as script_file:
        script_content = script_file.read()
    job_create_request = JobCreate(
        job=Job(name=name, script=script_content),
        user=user,
    )
    created_job_list = create_job([job_create_request])
    click.secho(f"Job successfully created!\n\n{created_job_list[0]}", fg="green")


@click.command()
@click.option("--job_id", required=True, help="id of the job")
def schedule(job_id) -> None:
    """Create cluster."""
    # fetch a user first
    user_access_token = config.USER_ACCESS_TOKEN
    auther = auther_registry.get_auther(config.AUTHER)
    user = auther.extract_user_from_token(user_access_token)

    fetched_jobs = get_jobs(
        JobGetQueryParams(query_type=JobQueryType.GET_JOBS_BY_ID, job_id=job_id)
    )
    if len(fetched_jobs) == 0:
        click.secho(f"Job with id {job_id} not found!", fg="red")
    else:
        schedule_create_request = ScheduleCreate(
            schedule=Schedule(job_id=job_id, user_id=user.user_id, job=fetched_jobs[0]),
            user=user,
        )
        created_schedule_list = create_schedule([schedule_create_request])
        click.secho(
            f"Schedule successfully created!\n\n{created_schedule_list[0]}", fg="green"
        )


@click.group()
def create() -> None:
    """Create resources."""
    pass


create.add_command(user)
create.add_command(cluster)
create.add_command(job)
create.add_command(schedule)



"""Blackcap celery tasks."""

from typing import Any

from celery import Celery

from flask import Flask

from blackcap.tasks.pub_messenger import publish_messenger  # noqa: F401


def init_celery(app: Flask, celery_app: Celery) -> None:
    """Initialize celery app.

    Args:
        app (Flask): Flask app
        celery_app (Celery): Celery app
    """
    celery_app.conf.update(app.config)

    class ContextTask(celery_app.Task):
        def __call__(self: "ContextTask", *args: Any, **kwargs: Any) -> Any:
            with app.app_context():
                return self.run(*args, **kwargs)

    celery_app.Task = ContextTask



"""Base Auther class."""

from abc import ABC, abstractclassmethod
from typing import List, Optional, Tuple

from blackcap.schemas.api.auth.post import AuthUserCreds
from blackcap.schemas.api.user.post import UserCreate
from blackcap.schemas.user import User

from flask import Request


class BaseAuther(ABC):
    """Base Auther class."""

    CONFIG_KEY = "AUTHER"
    CONFIG_KEY_DEF_VAL = "COOKIE"

    # Change this value in custom auther implementations.
    CONFIG_KEY_VAL = "COOKIE"

    @abstractclassmethod
    def register_user(
        self: "BaseAuther", user_create_list: List[UserCreate]
    ) -> List[User]:  # noqa: E501
        """Register user.

        Args:
            user_create_list (List[UserCreate]): List of users to register

        Raises:
            Exception: error  # noqa: DAR402

        Returns:
            List(User): List of registered users  # noqa: DAR202
        """
        pass

    @abstractclassmethod
    def login_user(
        self: "BaseAuther", user_creds: AuthUserCreds
    ) -> Optional[Tuple[User, str]]:
        """Login user.

        Args:
            user_creds (AuthUserCreds): user creds

        Raises:
            Exception: error  # noqa: DAR402

        Returns:
            Optional[Tuple[User, str]]: user and cookie tuple or None
        """
        pass

    @abstractclassmethod
    def logout_user(
        self: "BaseAuther",
    ) -> None:
        """Logout user."""
        pass

    @abstractclassmethod
    def extract_user_from_token(
        self: "BaseAuther", token: str
    ) -> Optional[User]:  # noqa: E501
        """Extract user from token.

        Args:
            token (str): user access token

        Raises:
            Exception: error  # noqa: DAR402

        Returns:
            Optional[User]: Instance of User or None
        """
        pass

    @abstractclassmethod
    def authorize_user(
        self: "BaseAuther", user: User, request: Request
    ) -> bool:  # noqa: E501
        """Authorize user actions on resources.

        Args:
            user (User): Instance of User
            request (Request): Instance of Flask Request

        Returns:
            bool: Authorization decision  # noqa: DAR202
        """
        pass



"""Blackcap user POST route."""

import json
from http import HTTPStatus
from typing import List

from blackcap.blocs.user import create_user
from blackcap.routes.user import user_bp
from blackcap.schemas.api.user.post import (
    UserCreate,
    UserPOSTRequest,
    UserPOSTResponse,
)

from flask import Response, make_response, request

from pydantic import ValidationError, parse_obj_as

from sqlalchemy.exc import SQLAlchemyError


@user_bp.post("/")
def post_user() -> Response:
    """Post User.

    Returns:
        Response: Flask response
    """
    # Parse json from request
    try:
        user_post_request = UserPOSTRequest.parse_obj(json.loads(request.json))
        user_create_list = parse_obj_as(
            List[UserCreate], user_post_request.users
        )  # noqa: E501
    except ValidationError as e:
        response_body = UserPOSTResponse(
            msg="json validation failed", errors=e.errors()
        )  # noqa: E501
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = UserPOSTResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # ceate user in DB

    try:
        user_list = create_user(user_create_list)
    except SQLAlchemyError:
        response_body = UserPOSTResponse(
            msg="internal databse error", errors=["internal database error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501
    except Exception:
        response_body = UserPOSTResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # return created users in response
    response_body = UserPOSTResponse(
        msg="users successfully created", items=user_list
    )  # noqa: E501
    return make_response(response_body.json(), HTTPStatus.OK)



"""Scheduler to schedule jobs using metrcis from clusters."""

from blackcap.scheduler.random_scheduler import RandomScheduler
from blackcap.scheduler.registry import SchedulerRegistry

scheduler_registry = SchedulerRegistry()
scheduler_registry.add_scheduler(RandomScheduler())



"""Auther registry."""

from typing import Optional

from blackcap.auther.base import BaseAuther


class AutherRegistry:
    """Auther registry."""

    authers = {}

    def add_auther(self: "AutherRegistry", auther: BaseAuther) -> None:
        """Add custom authers to registry.

        Args:
            auther (BaseAuther): Custom auther implementation
        """
        self.authers[auther.CONFIG_KEY_VAL] = auther

    def get_auther(
        self: "AutherRegistry", auther: str
    ) -> Optional[BaseAuther]:  # noqa: E501
        """Get auther.

        Args:
            auther (str): Auther name

        Returns:
            Optional[BaseAuther]: Returns the auther if found else None
        """
        return self.authers.get(auther)



"""Blackcap Cluster POST route schemas."""

from typing import List, Optional

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.cluster import Cluster
from blackcap.schemas.user import User

from pydantic import BaseModel


class ClusterCreate(BaseModel):
    """Schema to parse create Cluster requests."""

    cluster: Cluster
    user: Optional[User]


class ClusterPOSTRequest(BaseModel):
    """Cluster POST request schema."""

    Clusters: List[ClusterCreate]


class ClusterPOSTResponse(ResponseSchema):
    """Cluster POST response schema."""

    items: List[Cluster] = []



"""Schedule DBModel."""


from blackcap.models.meta.mixins import (
    DBModel,
    SurrogatePKUUID,
    TimestampMixin,
)  # noqa: E501
from blackcap.models.meta.orm import reference_col

from sqlalchemy import Column, DateTime, String

# from sqlalchemy.orm import relationship


class ScheduleDB(DBModel, TimestampMixin, SurrogatePKUUID):
    """Schedule table."""

    __tablename__ = "schedule"
    serialize_rules = (
        "-cluster.schedules",
        "-job.schedules",
        "-protagonist.schedules",
    )
    job_id = reference_col("job")
    # job = relationship("JobDB", backref="schedules")
    assigned_cluster_id = reference_col("cluster")
    # assigned_cluster = relationship("ClusterDB", backref="schedules")
    # TODO: this should be an Enum with the possible status values
    status = Column(String, nullable=False, default="PENDING")
    started_at = Column(DateTime(timezone=True), nullable=True)
    finished_at = Column(DateTime(timezone=True), nullable=True)
    protagonist_id = reference_col("protagonist")
    # protagonist = relationship("ProtagonistDB", backref="schedules")



"""Command-line interface."""

# flake8: noqa

import os

import click

from blackcap.cli.create import create
from blackcap.cli.db import db
from blackcap.cli.get import get
from blackcap.cli.publish import pub
from blackcap.cli.schedule import sched
from blackcap.cli.subscribe import sub
from blackcap.configs import config_registry
from blackcap.auther import auther_registry
from blackcap.schemas.api.auth.post import AuthUserCreds


from .. import __version__

config = config_registry.get_config()


@click.command()
@click.option("--email", required=True, help="email of user")
@click.option("--password", required=True, help="password of user")
def login(email, password) -> None:
    auth_creds = AuthUserCreds(email=email, password=password)
    auther = auther_registry.get_auther(config.AUTHER)
    login_tuple = auther.login_user(auth_creds)
    if login_tuple is None:
        click.secho("Login failed!", fg="red")
    else:
        click.secho(f"Your token: {login_tuple[1]}")
        click.secho(f"Add it to your env\n\n export USER_ACCESS_TOKEN={login_tuple[1]}")


@click.group()
@click.version_option(version=__version__)
def main() -> None:
    """Conductor console."""
    pass


main.add_command(create)
main.add_command(get)
main.add_command(db)
main.add_command(pub)
main.add_command(sched)
main.add_command(login)
main.add_command(sub)



"""publish commands."""
# flake8: noqa: DAR101

from datetime import datetime
import random
from pathlib import Path

import click

from blackcap.schemas.message import Message, MessageType
from blackcap.tasks.pub_messenger import publish_messenger

from logzero import logger


@click.command()
@click.option(
    "--topic", default="test-topic", help="Id of the topic to publish msgs on"
)
def chatter(topic: str) -> None:
    """Publish a random msg."""
    logger.info("Trying to publish a random msg...")
    try:
        msg = f"Your lucky number is: {random.randint(1, 10)}"
        msg = Message(
            msg_type=MessageType.TO_DEMON_SCHEDULE_MSG,
            data={"msg": msg},
            timestamp=str(datetime.now()),
        )
        publish_messenger.delay(msg.dict(), topic)
        logger.info(f"\nSuccess!\n\nMsg: {msg}")
    except Exception as e:
        logger.error(f"failed to publish msg: {e}")


@click.command()
@click.option(
    "--topic", default="test-topic", help="Id of the topic to publish msgs on"
)
@click.option("--data", help="custom string to publish", required=True)
def custom(topic: str, data: str) -> None:
    """Publish user provided string."""
    logger.info("Trying to publish a custom string...")
    try:
        msg = Message(
            msg_type=MessageType.TO_DEMON_SCHEDULE_MSG,
            data={"msg": data},
            timestamp=str(datetime.now()),
        )
        publish_messenger.delay(msg.dict(), topic)
        logger.info(f"\nSuccess!\n\nMsg: {data}")
    except Exception as e:
        logger.error(f"failed to publish msg: {e}")


@click.command()
@click.option(
    "--topic", default="test-topic", help="Id of the topic to publish msgs on"
)
@click.option("--file", help="file path of job's JSON file", required=True)
def job(topic: str, file: str) -> None:
    """Publish a job."""
    logger.info("Trying to publish a job...")
    try:
        job_file_path = Path(file)
        with open(job_file_path) as job_file:
            data = job_file.read()
            msg = Message(
                msg_type=MessageType.TO_DEMON_SCHEDULE_MSG,
                data={"msg": data},
                timestamp=datetime.now(),
            )
            publish_messenger.delay(msg.dict(), topic)
            logger.info(f"\nSuccess!\n\nMsg: {data}")
    except Exception as e:
        logger.error(f"failed to publish msg: {e}")


@click.group()
def pub() -> None:
    """GCP pub/sub channel publish commands."""
    pass


pub.add_command(chatter)
pub.add_command(custom)
pub.add_command(job)



"""Message schema."""

from enum import Enum, unique
from typing import Dict

from pydantic import BaseModel, validator


@unique
class MessageType(Enum):
    """Message types."""

    TO_DEMON_SCHEDULE_MSG = "to_demon_schedule_msg"
    TO_CONDUCTOR_JOB_STATUS_UPDATE = "to_conductor_job_status_update"


class Message(BaseModel):
    """Message schema to serialize/deserialze messenger msgs."""

    msg_type: MessageType
    data: Dict
    timestamp: str

    @validator("msg_type")
    def convert_msg_type_enum_to_str(cls: "Message", v: MessageType) -> str:
        """Convert enum to str.

        Args:
            v (MessageType): MessageType Enum

        Returns:
            str: string value of enum
        """
        return v.value



"""Default blackcap config."""

from blackcap.configs.base import BaseConfig


class DefaultConfig(BaseConfig):
    """Default blackcap config."""

    def get_config_name(self: "DefaultConfig") -> str:
        """Return Config name.

        Returns:
            str: Name of the config
        """
        return "DEFAULT"

    def get_sql_db_uri(self: "DefaultConfig") -> str:
        """Get uri of the sql database.

        Returns:
            str: uri of the sql database
        """
        return f"{self.DB_TYPE}://{self.DB_USER}:{self.DB_PASS}@{self.DB_HOST}:{self.DB_PORT}/{self.DB_NAME}"  # noqa: E501

    @property
    def SQLALCHEMY_DATABASE_URI(self: "DefaultConfig") -> str:
        """Get db uri for sqlalchemy.

        Returns:
            str: Database uri for sqlalchemy
        """
        if self.DB_URI is not None:
            return self.DB_URI
        else:
            return self.get_sql_db_uri()



"""Blackcap Cluster PUT route schemas."""

from typing import Optional

from pydantic import BaseModel
from pydantic.types import UUID4


class ClusterUpdate(BaseModel):
    """Schema to parse update cluster requests."""

    cluster_id: UUID4
    cluster_type: Optional[str]
    cluster_caps: Optional[str]
    messenger: Optional[str]
    messenger_queue: Optional[str]



"""Blackcap Cluster GET route schemas."""

from enum import Enum, unique
from typing import List

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.cluster import Cluster

from pydantic import BaseModel


@unique
class ClusterQueryType(Enum):
    """Cluster Query type enum."""

    GET_ALL_CLUSTER = "get_all_clusters"


class ClusterGetQueryParams(BaseModel):
    """Cluster GET request query params schema."""

    query_type: ClusterQueryType


class ClusterGetResponse(ResponseSchema):
    """Cluster GET response schema."""

    items: List[Cluster] = []



"""Blackcap Schedule POST route schemas."""

from typing import List, Optional

from blackcap.schemas.schedule import Schedule
from blackcap.schemas.user import User

from pydantic import BaseModel


class SchedulePOSTResponse(BaseModel):
    """Schedule POST response schema."""

    items: List[Schedule] = []


class ScheduleCreate(BaseModel):
    """Schema to parse create schedule requests."""

    schedule: Schedule
    user: Optional[User]


class SchedulePOSTRequest(BaseModel):
    """Schedule POST request schema."""

    schedules: List[ScheduleCreate]



"""Blackcap User DELETE route schemas."""

from pydantic import BaseModel
from pydantic.types import UUID4


class UserDelete(BaseModel):
    """Schema to parse delete user requests."""

    user_id: UUID4



"""Scheduler registry."""

from typing import Optional

from blackcap.scheduler.base import BaseScheduler


class SchedulerRegistry:
    """Scheduler registry."""

    schedulers = {}

    def add_scheduler(self: "SchedulerRegistry", scheduler: BaseScheduler) -> None:
        """Add custom schedulers to registry.

        Args:
            scheduler (BaseScheduler): Custom scheduler implementation
        """
        self.schedulers[scheduler.CONFIG_KEY_VAL] = scheduler

    def get_scheduler(
        self: "SchedulerRegistry", scheduler: str
    ) -> Optional[BaseScheduler]:  # noqa: E501
        """Get scheduler.

        Args:
            scheduler (str): Scheduler name

        Returns:
            Optional[BaseScheduler]: Returns the scheduler if found else None
        """
        return self.schedulers.get(scheduler)



"""Schedule schema."""

from typing import Optional

from blackcap.schemas.job import Job

from pydantic import BaseModel
from pydantic.types import UUID4


class Schedule(BaseModel):
    """Schedule schema."""

    schedule_id: Optional[UUID4]
    user_id: Optional[UUID4]
    job_id: UUID4
    job: Optional[Job]
    assigned_cluster_id: Optional[UUID4]
    messenger: Optional[str]
    messenger_queue: Optional[str]
    status: str = "PENDING"



"""Job schema."""

from typing import Optional

from pydantic import BaseModel
from pydantic.types import UUID4


class Job(BaseModel):
    """Job schema."""

    job_id: Optional[UUID4]
    name: str
    script: str



"""Blackcap schedule POST route."""

from http import HTTPStatus

from blackcap.blocs.schedule import create_schedule
from blackcap.routes.job import job_bp
from blackcap.schemas.api.job.post import JobCreate, JobPOSTResponse

from flask import Response, make_response, request

from pydantic import ValidationError

from sqlalchemy.exc import SQLAlchemyError


@job_bp.post("/")
def post_job() -> Response:
    """Post job.

    Returns:
        Response: Flask response
    """
    # Parse json from request
    try:
        job_create = JobCreate.parse_obj(request.json)
    except ValidationError as e:
        response_body = JobPOSTResponse(
            msg="json validation failed", errors=e.errors()
        )  # noqa: E501
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = JobPOSTResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # ceate job in DB and publish msg
    try:
        job = create_schedule(job_create)
    except SQLAlchemyError:
        response_body = JobPOSTResponse(
            msg="internal databse error", errors=["internal database error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501
    except Exception:
        response_body = JobPOSTResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # return created job in response
    response_body = JobPOSTResponse(
        msg="job successfully created", items=[job.dict()]
    )  # noqa: E501
    return make_response(response_body.json(), HTTPStatus.OK)



"""Config registry."""

from typing import Optional
import os

from blackcap.configs.base import BaseConfig, ConfigOfConfig


class ConfigRegistry:
    """Config registry."""

    configs = {}
    config_of_config = ConfigOfConfig()

    def add_config(self: "ConfigRegistry", config: BaseConfig) -> None:
        """Add custom configs to registry.

        Args:
            config (BaseConfig): Custom config implementation
        """
        self.configs[config.get_config_name()] = config

    def get_config(
        self: "ConfigRegistry", config: str = config_of_config.BLACKCAP_CONFIG
    ) -> Optional[BaseConfig]:  # noqa: E501
        """Get config.

        Args:
            config (str): Config name

        Returns:
            Optional[BaseConfig]: Returns the config if found else None
        """
        return self.configs.get(config)



"""Blackcap Schedule PUT route schemas."""

from pydantic import BaseModel
from pydantic.types import UUID4


class ScheduleUpdate(BaseModel):
    """Schema to parse update schedule requests."""

    schedule_id: UUID4



"""Blackcap job routes."""

from flask import Blueprint


job_bp = Blueprint("job", __name__, url_prefix="/v1/job")


from blackcap.routes.job.get import get_job  # noqa: F401, E402, I100
from blackcap.routes.job.post import post_job  # noqa: F401, E402, I100



"""Blackcap schedule GET route."""

from http import HTTPStatus

from blackcap.blocs.schedule import get_schedules
from blackcap.routes.schedule import schedule_bp
from blackcap.schemas.api.schedule.get import (
    ScheduleGetQueryParams,
    ScheduleGetResponse,
)

from flask import Response, make_response, request

from pydantic.error_wrappers import ValidationError

from sqlalchemy.exc import SQLAlchemyError


@schedule_bp.get("/")
def get_schedule() -> Response:
    """Get schedule.

    Returns:
        Response: Flask response
    """
    # Parse query params from request
    try:
        query_params = ScheduleGetQueryParams.parse_raw(request.query_string)
    except ValidationError as e:
        response_body = ScheduleGetResponse(
            msg="query validation error", errors=e.errors()
        )  # noqa: E501
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = ScheduleGetResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # Get s from the DB
    try:
        schedules = get_schedules(query_params)
    except SQLAlchemyError:
        response_body = ScheduleGetResponse(
            msg="internal databse error", errors=["internal database error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501
    except Exception:
        response_body = ScheduleGetResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # return fetched s in response
    response_body = ScheduleGetResponse(
        msg=" successfully retrieved", items=schedules
    )  # noqa: E501
    return make_response(response_body.json(), HTTPStatus.OK)



"""Mixins for database models."""

import uuid
from datetime import datetime
from typing import Any, List, Optional, Union

from blackcap.models.meta.helpers import GUID

from sqlalchemy import Column, DateTime, Integer
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm.session import Session

from sqlalchemy_serializer import SerializerMixin


# declarative base class
Base = declarative_base()


class TimestampMixin(object):
    """Mixin that adds timestamp."""

    created_at = Column(
        DateTime(timezone=True), nullable=False, default=datetime.utcnow
    )
    updated_at = Column(DateTime(timezone=True), onupdate=datetime.utcnow)


class CRUDMixin(object):
    """Mixin that adds convenience methods for CRUD operations."""

    @classmethod
    def create(cls: Any, DBSession: Session, **kwargs: Any) -> Any:
        """Create a new record and save it to the database.

        Args:
            DBSession: (Session): database session to use
            **kwargs (Any): kwargs to pass to cls for init

        Returns:
            Any: Instance of self
        """
        instance = cls(**kwargs)
        return instance.save()

    @classmethod
    def bulk_create(cls: Any, obj_list: List[Any], DBSession: Session) -> Any:
        """Create a new records and save them to the database.

        Args:
            list (List[Any]): List of self
            DBSession: (Session): database session to use

        Returns:
            Any: True if scuccess
        """
        DBSession.bulk_save_objects(obj_list, return_defaults=True)
        DBSession.commit()
        return True

    def update(
        self: "CRUDMixin",
        DBSession: Session,
        commit: bool = True,
        **kwargs: Any  # noqa: E501
    ) -> Any:
        """Update specific fields of a record.

        Args:
            DBSession: (Session): database session to use
            commit (bool): Flag to control commit behaviour. Defaults to True.
            **kwargs (Any): kwargs to update cls params

        Returns:
            Any: Instance of self
        """
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        return commit and self.save(DBSession) or self

    @classmethod
    def bulk_update(cls: Any, list: List[Any], DBSession: Session) -> Any:
        """Create a new records and save them to the database.

        Args:
            list (List[Any]): List of instances of self
            DBSession: (Session): database session to use

        Returns:
            Any: List of updated instances self
        """
        instance_list: List[Any] = []
        for item in list:
            instance_list.append(cls(**item))

        DBSession.bulk_update_mappings(list)
        DBSession.commit()
        return instance_list

    def save(
        self: "CRUDMixin", DBSession: Session, commit: bool = True
    ) -> Any:  # noqa: E501
        """Save the record.

        Args:
            DBSession: (Session): database session to use
            commit (bool): Flag to control commit behaviour. Defaults to True. # noqa: E501

        Returns:
            Any: Instance of self
        """
        DBSession.add(self)
        if commit:
            DBSession.commit()
        return self

    def delete(
        self: "CRUDMixin", DBSession: Session, commit: bool = True
    ) -> Union[bool, Any]:
        """Remove the record from the database.

        Args:
            DBSession: (Session): database session to use
            commit (bool): Flag to control commit behaviour. Defaults to True. # noqa: E501

        Returns:
            Union[bool, Any]: True if commit is successful
        """
        DBSession.delete(self)
        return commit and DBSession.commit()


class DBModel(CRUDMixin, SerializerMixin, Base):
    """Base model class that includes CRUD convenience methods."""

    __abstract__ = True


# From Mike Bayer's "Building the app" talk
# https://speakerdeck.com/zzzeek/building-the-app
class SurrogatePK(object):
    """Adds a surrogate int 'primary key' column named ``id`` to any orm class."""  # noqa: E501

    __table_args__ = {"extend_existing": True}

    id = Column(Integer, primary_key=True)

    @classmethod
    def get_by_id(cls: Any, record_id: Union[str, int]) -> Union[None, Any]:
        """Get record by ID.

        Args:
            record_id (Union[str, int]): Id of the record to fetch

        Returns:
            Union[None, Any]: Return the instance of self if found or else None
        """
        if any(
            (
                isinstance(record_id, str) and record_id.isdigit(),
                isinstance(record_id, (int, float)),
            )
        ):
            return cls.query.get(int(record_id))
        return None


class SurrogatePKUUID(object):
    """Adds a surrogate int 'primary key' column named ``id`` to any orm class."""  # noqa: E501

    __table_args__ = {"extend_existing": True}

    id = Column(GUID, primary_key=True, default=lambda: str(uuid.uuid4()))

    @classmethod
    def get_by_uuid(
        cls: Any, record_id: Union[str, uuid.UUID]
    ) -> Optional[Any]:  # noqa: E501
        """Get record by ID.

        Args:
            record_id (Union[str, uuid.UUID]): Id of the record to fetch

        Returns:
            Union[None, Any]: Return the instance of self if found or else None
        """
        if any((isinstance(record_id, str), isinstance(record_id, uuid.UUID))):
            return cls.query.get(uuid.UUID(str(record_id)))
        return None



"""Blackcap schedule routes."""


from flask import Blueprint

schedule_bp = Blueprint("schedule", __name__, url_prefix="/v1/schedule")

from conductor.routes.schedule.get import get_schedule  # noqa: F401, E402, I100, E501

# from conductor.routes.schedule.post import post_schedule  # noqa: F401, E402, I100, E501



"""Blackcap database."""

from blackcap.configs import config_registry

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

db_engine = create_engine(config_registry.get_config().SQLALCHEMY_DATABASE_URI)
# Use this session for all db operations in the app
DBSession = sessionmaker(db_engine)



"""Cluster BLoCs."""

from typing import List

from blackcap.db import DBSession
from blackcap.models.cluster import ClusterDB
from blackcap.schemas.api.cluster.get import (
    ClusterGetQueryParams,
    ClusterQueryType,
)  # noqa: E501
from blackcap.schemas.api.cluster.post import ClusterCreate
from blackcap.schemas.cluster import Cluster

from logzero import logger

from sqlalchemy import select


def create_cluster(
    cluster_create_list: List[ClusterCreate],
) -> List[Cluster]:  # noqa: E501
    """Create clusters from create cluster request.

    Args:
        cluster_create_list (List[ClusterCreate]): ClusterCreate request

    Raises:
        Exception: error

    Returns:
        List[Cluster]: List of created clusters
    """
    with DBSession() as session:
        try:
            cluster_db_create_list: List[ClusterDB] = [
                ClusterDB(
                    **cluster_create.cluster.dict(exclude={"cluster_id"})
                )  # noqa: E501
                for cluster_create in cluster_create_list
            ]
            ClusterDB.bulk_create(cluster_db_create_list, session)
            return [
                Cluster(cluster_id=obj.id, **obj.to_dict())
                for obj in cluster_db_create_list
            ]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to create clusters: {e}")
            raise e


def get_clusters(query_params: ClusterGetQueryParams) -> List[Cluster]:
    """Query DB for clusters.

    Args:
        query_params (ClusterGetQueryParams): Query params from request

    Raises:
        Exception: error

    Returns:
        List[Cluster]: List of clusters returned from DB
    """
    cluster_list: List[ClusterDB] = []

    stmt = ""

    if query_params.query_type == ClusterQueryType.GET_ALL_CLUSTER:
        stmt = select(ClusterDB)

    with DBSession() as session:
        try:
            cluster_list: List[ClusterDB] = (
                session.execute(stmt).scalars().all()
            )  # noqa: E501
            return [
                Cluster(cluster_id=obj.id, **obj.to_dict())
                for obj in cluster_list  # noqa: E501
            ]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to fetch clusters due to {e}")
            raise e



"""Blackcap Job GET route schemas."""

from enum import Enum, unique
from typing import List, Optional

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.job import Job

from pydantic import BaseModel


class JobGetResponse(ResponseSchema):
    """Job GET response schema."""

    items: List[Job] = []


@unique
class JobQueryType(Enum):
    """Job Query type enum."""

    GET_ALL_JOBS = "get_all_jobs"
    GET_JOBS_BY_ID = "get_job_by_id"
    GET_JOBS_BY_PROTAGONIST_ID = "get_jobs_by_protagonist_id"
    GET_JOBS_BY_CLUSTER_ID = "get_jobs_by_cluster_id"
    GET_JOBS_BY_CREATE_TIMERANGE = "get_jobs_by_create_timerange"
    GET_JOBS_BY_FINISHED_TIMERANGE = "get_jobs_by_finished_timerange"
    GET_JOBS_BY_STATUS = "get_jobs_by_status"


class JobGetQueryParams(BaseModel):
    """Job GET request query params schema."""

    query_type: JobQueryType
    job_id: Optional[str]
    protagonist_id: Optional[str]
    cluster_id: Optional[str]
    create_timerange: Optional[str]
    finished_timerange: Optional[str]
    job_status: Optional[str]



"""Random Scheduler implementation of Scheduler."""

import random
from typing import List

from blackcap.db import DBSession
from blackcap.models.cluster import ClusterDB
from blackcap.models.schedule import ScheduleDB
from blackcap.scheduler.base import BaseScheduler
from blackcap.schemas.api.schedule.post import ScheduleCreate
from blackcap.schemas.schedule import Schedule

from logzero import logger

from sqlalchemy.sql.expression import select


class RandomScheduler(BaseScheduler):
    """Random scheduler schedule jobs randomly."""

    CONFIG_KEY_VAL = "RANDOM"

    def schedule(
        self: "BaseScheduler", schedule_create: ScheduleCreate
    ) -> Schedule:  # noqa: E501
        """Create schedule from schedule request.

        Args:
            schedule_create (ScheduleCreate): Schedule create request

        Raises:
            Exception: No cluster available

        Returns:
            Schedule: Instance of Created Schedule
        """
        # fetch available clusters
        cluster_list: List[ClusterDB] = []
        with DBSession() as session:
            try:
                stmt = select(ClusterDB)
                cluster_list = session.execute(stmt).scalars().all()  # noqa: E501
            except Exception as e:
                logger.error(f"Unable to fetch clusters: {e}")
                raise e

        # Raise error if no cluster available
        if len(cluster_list) == 0:
            logger.error("No cluster available")
            raise Exception("cluster_list: List[ClusterDB]")

        # Randomly selects a cluster
        selected_cluster = random.choice(cluster_list)  # noqa: S311

        schedule_create.schedule.assigned_cluster_id = selected_cluster.id

        # Add schedule to DB
        schedule_dict = schedule_create.schedule.dict(
            exclude={"schedule_id", "messenger", "messenger_queue", "job"}
        )
        user_id = schedule_dict.pop("user_id")
        try:
            schedule = ScheduleDB(
                protagonist_id=user_id,
                **schedule_dict,
            )
            schedule.save(session)
            return Schedule(
                schedule_id=schedule.id,
                user_id=user_id,
                messenger=selected_cluster.messenger,
                messenger_queue=selected_cluster.messenger_queue,
                job=schedule_create.schedule.job,
                **schedule.to_dict(),
            )
        except Exception as e:
            logger.error(f"Unable to save schedule to DB: {e}")
            raise e



"""Schedule BLoCs."""

from datetime import datetime
from typing import List

from blackcap.configs import config_registry
from blackcap.db import DBSession
from blackcap.scheduler import scheduler_registry
from blackcap.models.schedule import ScheduleDB
from blackcap.schemas.api.schedule.delete import ScheduleDelete
from blackcap.schemas.api.schedule.get import (
    ScheduleGetQueryParams,
    ScheduleQueryType,
)  # noqa: E501
from blackcap.schemas.api.schedule.post import ScheduleCreate
from blackcap.schemas.api.schedule.put import ScheduleUpdate
from blackcap.schemas.message import Message, MessageType
from blackcap.schemas.schedule import Schedule
from blackcap.tasks.pub_messenger import publish_messenger

from logzero import logger

from sqlalchemy import select


config = config_registry.get_config()
scheduler = scheduler_registry.get_scheduler(config.SCHEDULER)


def create_schedule(
    schedule_create_list: List[ScheduleCreate],
) -> List[Schedule]:  # noqa: E501
    """Create schedules and save to DB from ScheduleCreate requests.

    Args:
        schedule_create_list (List[ScheduleCreate]): List of ScheduleCreate request  # noqa: E501

    Returns:
        List[Schedule]: Instance of Schedule from DB
    """
    schedule_list: List[Schedule] = [
        scheduler.schedule(schedule_create)
        for schedule_create in schedule_create_list  # noqa: E501
    ]

    # publish msg to demon
    for schedule in schedule_list:
        message = Message(
            msg_type=MessageType.TO_DEMON_SCHEDULE_MSG,
            data=schedule.dict(),
            timestamp=str(datetime.now()),
        )
        publish_messenger.delay(message.dict(), schedule.messenger_queue)

    return schedule_list


def get_schedules(query_params: ScheduleGetQueryParams) -> List[ScheduleDB]:
    """Query DB for schedules.

    Args:
        query_params (ScheduleGetQueryParams): Query params from request

    Raises:
        Exception: error

    Returns:
        List[Schedule]: List of schedules returned from DB
    """
    schedule_list: List[ScheduleDB]
    if query_params.query_type == ScheduleQueryType.GET_ALL_SCHEDULES:
        stmt = select(ScheduleDB)
    if query_params.query_type == ScheduleQueryType.GET_SCHEDULE_BY_ID:
        stmt = select(ScheduleDB).where(
            ScheduleDB.id == query_params.schedule_id
        )  # noqa: E501
    if (
        query_params.query_type
        == ScheduleQueryType.GET_SCHEDULES_BY_PROTAGONIST_ID  # noqa: E501
    ):
        stmt = select(ScheduleDB).where(
            ScheduleDB.protagonist_id == query_params.protagonist_id
        )  # noqa: E501
    if (
        query_params.query_type
        == ScheduleQueryType.GET_SCHEDULES_BY_CLUSTER_ID  # noqa: E501
    ):
        stmt = select(ScheduleDB).where(
            ScheduleDB.assigned_cluster_id == query_params.cluster_id
        )  # noqa: E501

    with DBSession() as session:
        try:
            schedule_list: List[ScheduleDB] = (
                session.execute(stmt).scalars().all()
            )  # noqa: E501
            return schedule_list
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to fetch schedules due to {e}")
            raise e

    return schedule_list


def update_schedule(schedule_update: ScheduleUpdate) -> ScheduleDB:
    """Update schedule in the DB from ScheduleUpdate request.

    Args:
        schedule_update (ScheduleUpdate): ScheduleUpdate request

    Raises:
        Exception: error

    Returns:
        ScheduleDB: Instance of Updated Schedule
    """
    stmt = select(ScheduleDB).where(
        ScheduleDB.id == schedule_update.schedule_id
    )  # noqa: E501
    with DBSession() as session:
        try:
            schedule_list: List[ScheduleDB] = (
                session.execute(stmt).scalars().all()
            )  # noqa: E501
            # if schedule_list is None:
            # raise NotFound exception - 404
            if len(schedule_list) == 1:
                schedule_update_dict = schedule_update.dict()
                schedule_update_dict.pop("schedule_id")
                updated_schedule = schedule_list[0].update(
                    session, **schedule_update.dict()
                )  # noqa: E501
                return updated_schedule
        except Exception as e:
            session.rollback()
            logger.error(
                f"Unable to update schedule: {schedule_update.dict()} due to {e}"  # noqa: E501
            )
            raise e


def delete_schedule(schedule_delete: ScheduleDelete) -> ScheduleDB:
    """Delete schedule in the DB from ScheduleDelete request.

    Args:
        schedule_delete (ScheduleDelete): ScheduleDelete request

    Raises:
        Exception: error

    Returns:
        ScheduleDB: Instance of Deleted Schedule
    """
    stmt = select(ScheduleDB).where(
        ScheduleDB.id == schedule_delete.schedule_id
    )  # noqa: E501
    with DBSession() as session:
        try:
            schedule_list: List[ScheduleDB] = (
                session.execute(stmt).scalars().all()
            )  # noqa: E501
            if not schedule_list:
                # TODO: raise not found exception
                pass
            if len(schedule_list) == 1:
                deleted_schedule = schedule_list[0].delete(session)
                return deleted_schedule
        except Exception as e:
            session.rollback()
            logger.error(
                f"Unable to delete schedule: {schedule_delete.dict()} due to {e}"  # noqa: E501
            )
            raise e



"""Blackcap DB commands."""

import click

from blackcap.db import db_engine
from blackcap.models.meta.mixins import DBModel


@click.command()
def init() -> None:
    """Init database."""
    click.secho("Creating database!", fg="green")
    DBModel.metadata.create_all(db_engine)


@click.command()
def destroy() -> None:
    """Destroy database."""
    click.secho("Destroying database!", fg="red")
    DBModel.metadata.drop_all(db_engine)


@click.command()
def reset() -> None:
    """Reset database."""
    DBModel.metadata.drop_all(db_engine)
    DBModel.metadata.create_all(db_engine)


@click.group()
def db() -> None:
    """Database commands."""
    pass


db.add_command(init)
db.add_command(destroy)
db.add_command(reset)



"""Messenger that delivers and reads msgs from a pub/sub queue."""

from blackcap.configs import config_registry
from blackcap.messenger.gcp_messenger import GCPMessenger
from blackcap.messenger.registry import MessengerRegistry

config = config_registry.get_config()

messenger_registry = MessengerRegistry()
messenger_registry.add_messenger(GCPMessenger(config))



"""JSON Encoders."""

import json
from typing import Any
from uuid import UUID


class UUIDEncoder(json.JSONEncoder):
    """UUID Json encoder."""

    def default(self: "UUIDEncoder", obj: Any) -> Any:
        """Convert UUID to str.

        Args:
            obj (Any): Object

        Returns:
            Any: [description]
        """
        if isinstance(obj, UUID):
            return str(obj)
        return json.JSONEncoder.default(self, obj)



"""Status route."""

from flask import Blueprint
from flask import Response

status_bp = Blueprint("status", __name__, url_prefix="/v1/status")


@status_bp.get("/")
def status_ok() -> Response:
    """Status route.

    Returns:
        Response: Flask response
    """
    return "OK"



"""Cluster schema."""

from typing import Optional

from pydantic import BaseModel
from pydantic.types import UUID4


class Cluster(BaseModel):
    """Cluster schema."""

    cluster_id: Optional[UUID4]
    name: str
    cluster_type: str
    status: str = "ONLINE"
    cluster_caps: Optional[str]
    messenger: str
    messenger_queue: str



"""Blackcap Auth PUT route schemas."""
from typing import List

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.user import User

from pydantic import BaseModel


class AuthCredsUpdate(BaseModel):
    """Schema to parse update user creds requests."""

    update_token: str
    new_password: str


class AuthPOSTResponse(ResponseSchema):
    """User POST response schema."""

    items: List[User]



"""Base Observer class."""

from abc import ABC, abstractclassmethod
from typing import List

from blackcap.schemas.metrics import Metrics


class BaseObserver(ABC):
    """Base Observer class."""

    CONFIG_KEY = "OBSERVER"
    CONFIG_KEY_DEF_VAL = "ELASTIC"

    # Change this value in custom auther implementations.
    CONFIG_KEY_VAL = "ELASTIC"

    @abstractclassmethod
    def get_metrics(self: "BaseObserver", range: str) -> List[Metrics]:  # noqa: E501
        """Get cluster metrics.

        Args:
            range (str): time range

        Returns:
            List[Metrics]: List of metrics from each cluster # noqa: DAR202
        """
        pass



"""Blackcap Schedule GET route schemas."""

from enum import Enum, unique
from typing import List, Optional

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.schedule import Schedule

from pydantic import BaseModel


class ScheduleGetQueryParams(BaseModel):
    """Schedule GET request query params schema."""

    schedule_id: Optional[str]
    job_id: Optional[str]
    protagonist_id: Optional[str]
    cluster_id: Optional[str]
    create_timerange: Optional[str]
    finished_timerange: Optional[str]


class ScheduleGetResponse(ResponseSchema):
    """Schedule GET response schema."""

    items: List[Schedule] = []


@unique
class ScheduleQueryType(Enum):
    """Schedule Query type enum."""

    GET_ALL_SCHEDULES = "get_all_schedules"
    GET_SCHEDULE_BY_ID = "get_schedule_by_id"
    GET_SCHEDULES_BY_PROTAGONIST_ID = "get_schedules_by_protagonist_id"
    GET_SCHEDULES_BY_CLUSTER_ID = "get_schedules_by_cluster_id"
    GET_SCHEDULES_BY_CREATE_TIMERANGE = "get_schedules_by_create_timerange"
    GET_SCHEDULES_BY_FINISHED_TIMERANGE = "get_schedules_by_finished_timerange"



"""Blackcap user GET route."""

from http import HTTPStatus

from blackcap.blocs.user import get_users
from blackcap.routes.user import user_bp
from blackcap.schemas.api.user.get import UserGetQueryParams, UserGetResponse

from flask import Response, make_response, request

from pydantic.error_wrappers import ValidationError

from sqlalchemy.exc import SQLAlchemyError


@user_bp.get("/")
def get_user() -> Response:
    """Get user.

    Returns:
        Response: Flask response
    """
    # Parse query params from request
    try:
        query_params = UserGetQueryParams.parse_obj(request.args)
    except ValidationError as e:
        response_body = UserGetResponse(
            msg="query validation error", errors=e.errors()
        )  # noqa: E501
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = UserGetResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # Get users from the DB
    try:
        user_list = get_users(query_params)
    except SQLAlchemyError:
        response_body = UserGetResponse(
            msg="internal databse error", errors=["internal database error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501
    except Exception:
        response_body = UserGetResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # return fetched users in response
    response_body = UserGetResponse(
        msg=" successfully retrieved", items=user_list
    )  # noqa: E501
    return make_response(response_body.json(), HTTPStatus.OK)



"""SQL helpers."""

import uuid
from typing import Any

from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.types import CHAR, TypeDecorator


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.
    Credit: https://gist.github.com/gmolveau/7caeeefe637679005a7bb9ae1b5e421e
    TODO: Change Any to more concrete types
    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self: "GUID", dialect: Any) -> Any:
        """Return column type to create based on dialect.

        Args:
            dialect (Any): sql dialect

        Returns:
            Any: column type to create
        """
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self: "GUID", value: Any, dialect: Any) -> Any:
        """Receive a bound parameter value to be converted.

        Args:
            value (Any): [description]
            dialect (Any): [description]

        Returns:
            Any: [description]
        """
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self: "GUID", value: Any, dialect: Any) -> Any:
        """Receive a result-row column value to be converted.

        Args:
            value (Any): [description]
            dialect (Any): [description]

        Returns:
            Any: [description]
        """
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value



"""Base config class, enum and helper function."""

from abc import ABC, abstractmethod
from typing import Optional

from pydantic import BaseSettings

from xdg import xdg_config_home


class BaseConfig(ABC, BaseSettings):
    """Base config class."""

    HOME: str
    FLASK_APP: str = "conductor"
    FLASK_HOST: str = "localhost"
    FLASK_PORT: int = 9991
    FLASK_ENV: str = "development"
    DB_TYPE: str = "postgresql"
    DB_NAME: str = "conductor"
    DB_URI: str = None
    DB_USER: str = "postgres"
    DB_PASS: str = "postgres"
    DB_HOST: str = "localhost"
    DB_PORT: int = 5432
    SQLALCHEMY_TRACK_MODIFICATIONS: bool = False
    SECRET_KEY: str = "ADD_A_RANDOM_KEY_HERE"  # noqa: S105
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 1440
    USER_ACCESS_TOKEN: Optional[str]
    AUTHER: str = "COOKIE"
    MESSENGER: str = "GCP"
    SCHEDULER: str = "RANDOM"
    OBSERVER: str = "ELASTIC"
    GOOGLE_APPLICATION_CREDENTIALS: str = "./keys.json"
    GCP_PROJECT_ID: str = "YOUR_GCP_PROJECT_ID"
    GCP_PUBSUB_TOPIC: str = "test-topic"
    GCP_PUBSUB_SUB_ID: str = "test-sub"
    CELERY_BROKER_URL: str = "redis://localhost:6379/0"
    CELERY_RESULT_BACKEND: str = "redis://localhost:6379/0"

    @abstractmethod
    def get_config_name(self: "BaseConfig") -> str:
        """Return Config name.

        Returns:
            str: Name of the config
        """
        pass

    @abstractmethod
    def get_sql_db_uri(self: "BaseConfig") -> str:
        """Get uri of the sql database.

        Returns:
            str: uri of the sql database
        """
        pass

    class Config:
        """Config for the BaseConfig class."""

        env_file = xdg_config_home() / ("orchestra") / ("blackcap.env")


class ConfigOfConfig(BaseSettings):
    "Config for loading config"

    BLACKCAP_CONFIG: str = "DEFAULT"

    class Config:
        """Config for ConfigofConfig class."""

        env_file = xdg_config_home() / ("orchestra") / ("blackcap.env")



"""Blackcap models."""

from blackcap.models.cluster import ClusterDB  # noqa: F401
from blackcap.models.job import JobDB  # noqa: F401
from blackcap.models.protagonist import ProtagonistDB  # noqa: F401
from blackcap.models.schedule import ScheduleDB  # noqa: F401



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['blackcap',
 'blackcap.auther',
 'blackcap.blocs',
 'blackcap.cli',
 'blackcap.configs',
 'blackcap.messenger',
 'blackcap.migrations',
 'blackcap.models',
 'blackcap.models.meta',
 'blackcap.observer',
 'blackcap.routes',
 'blackcap.routes.job',
 'blackcap.routes.schedule',
 'blackcap.routes.user',
 'blackcap.scheduler',
 'blackcap.schemas',
 'blackcap.schemas.api',
 'blackcap.schemas.api.auth',
 'blackcap.schemas.api.cluster',
 'blackcap.schemas.api.job',
 'blackcap.schemas.api.schedule',
 'blackcap.schemas.api.user',
 'blackcap.tasks',
 'blackcap.utils']

package_data = \
{'': ['*']}

install_requires = \
['Flask-Cors>=3.0.10,<4.0.0',
 'Flask>=2.0.1,<3.0.0',
 'PyJWT>=2.1.0,<3.0.0',
 'SQLAlchemy-serializer>=1.4.1,<2.0.0',
 'SQLAlchemy>=1.4.22,<2.0.0',
 'alembic>=1.6.5,<2.0.0',
 'bcrypt>=3.2.0,<4.0.0',
 'celery>=5.1.0,<6.0.0',
 'click<8.0',
 'google-cloud-pubsub>=2.5.0,<3.0.0',
 'gunicorn>=20.1.0,<21.0.0',
 'logzero>=1.7.0,<2.0.0',
 'psycopg2-binary>=2.8.6,<3.0.0',
 'pydantic[dotenv]>=1.8.2,<2.0.0',
 'redis>=3.5.3,<4.0.0',
 'xdg>=5.0.2,<6.0.0']

entry_points = \
{'console_scripts': ['blackcap = blackcap.cli.main:main']}

setup_kwargs = {
    'name': 'blackcap',
    'version': '0.6.0',
    'description': 'Shared library for Orchestra',
    'long_description': None,
    'author': 'Ankur Kumar',
    'author_email': 'ank@leoank.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""User schema."""

from typing import Optional

from pydantic import BaseModel
from pydantic.types import UUID4


class User(BaseModel):
    """User schema."""

    user_id: Optional[UUID4]
    name: str
    organisation: str
    email: str
    active: bool = True



"""Blackcap User PUT route schemas."""

from typing import Optional

from pydantic import BaseModel
from pydantic.types import UUID4


class UserUpdate(BaseModel):
    """Schema to parse update user requests."""

    user_id: UUID4
    name: Optional[str]
    email: Optional[str]
    organisation: Optional[str]



"""Messenger registry."""

from typing import Optional

from blackcap.messenger.base import BaseMessenger


class MessengerRegistry:
    """Messenger registry."""

    messengers = {}

    def add_messenger(self: "MessengerRegistry", messenger: BaseMessenger) -> None:
        """Add custom messengers to registry.

        Args:
            messenger (BaseMessenger): Custom messenger implementation
        """
        self.messengers[messenger.CONFIG_KEY_VAL] = messenger

    def get_messenger(
        self: "MessengerRegistry", messenger: str
    ) -> Optional[BaseMessenger]:  # noqa: E501
        """Get messenger.

        Args:
            messenger (str): Messenger name

        Returns:
            Optional[BaseMessenger]: Returns the messenger if found else None
        """
        return self.messengers.get(messenger)



"""Blackcap get commands."""
# flake8: noqa

import click

from blackcap.blocs.user import get_users
from blackcap.schemas.api.user.get import UserGetQueryParams, UserQueryType
from blackcap.schemas.user import User


@click.command()
@click.option("--type", required=True, help="type of the user query")
@click.option("--email", help="email of the user")
@click.option("--password", help="password of the user")
@click.option("--name", help="name of the user")
@click.option("--org", help="organisation of the user")
def user(type, email, password, name, org) -> None:
    """Get a user"""
    query_param = UserGetQueryParams(
        query_type=UserQueryType(type), name=name, email=email, organisation=org
    )
    fetched_user_list = get_users(query_param)
    click.secho(f"Users successfully fetched!\n\n", fg="green")
    for user in fetched_user_list:
        click.secho(f"{str(user.id)}, {user.name}, {user.email}, {user.organisation}")


@click.group()
def get() -> None:
    """Get resources."""
    pass


get.add_command(user)



"""Observer registry."""

from typing import Optional

from blackcap.observer.base import BaseObserver


class ObserverRegistry:
    """Observer registry."""

    observers = {}

    def add_observer(self: "ObserverRegistry", observer: BaseObserver) -> None:
        """Add custom observers to registry.

        Args:
            observer (BaseObserver): Custom observer implementation
        """
        self.observers[observer.CONFIG_KEY_VAL] = observer

    def get_observer(
        self: "ObserverRegistry", observer: str
    ) -> Optional[BaseObserver]:  # noqa: E501
        """Get observer.

        Args:
            observer (str): Observer name

        Returns:
            Optional[BaseObserver]: Returns the observer if found else None
        """
        return self.observers.get(observer)



"""Task to publish msgs to GCP Pub/Sub topic."""

from blackcap.configs import config_registry
from blackcap.messenger import messenger_registry
from blackcap.schemas.message import Message
from blackcap.workers import celery_app

from logzero import logger

config = config_registry.get_config()
messenger = messenger_registry.get_messenger(config.MESSENGER)


@celery_app.task
def publish_messenger(msg: Message, topic_id: str) -> None:
    """Publish msgs to GCP Pub/Sub topic.

    Args:
        msg (Message): msg to publish
        topic_id (str): Id of the topic

    Raises:
        Exception: error
    """
    try:
        messenger.publish(msg=msg, topic_id=topic_id)
    except Exception as e:
        logger.error(e)
        raise e



"""Blackcap job GET route."""

from http import HTTPStatus

from blackcap.blocs.job import get_jobs
from blackcap.routes.job import job_bp
from blackcap.schemas.api.job.get import JobGetQueryParams, JobGetResponse

from flask import Response, make_response, request

from pydantic.error_wrappers import ValidationError

from sqlalchemy.exc import SQLAlchemyError


@job_bp.get("/")
def get_job() -> Response:
    """Get job.

    Returns:
        Response: Flask response
    """
    # Parse query params from request
    try:
        query_params = JobGetQueryParams.parse_obj(request.args)
    except ValidationError as e:
        response_body = JobGetResponse(
            msg="query validation error", errors=e.errors()
        )  # noqa: E501
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = JobGetResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # Get jobs from the DB
    try:
        jobs = get_jobs(query_params)
    except SQLAlchemyError:
        response_body = JobGetResponse(
            msg="internal databse error", errors=["internal database error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501
    except Exception:
        response_body = JobGetResponse(
            msg="unknown error", errors=["unknown internal error"]
        )
        return make_response(
            response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR
        )  # noqa: E501

    # return fetched jobs in response
    response_body = JobGetResponse(
        msg="job successfully retrieved", items=jobs
    )  # noqa: E501
    return make_response(response_body.json(), HTTPStatus.OK)
